#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division
from .mib_db_api import *
from pyvoltha.adapters.extensions.omci.database.mib_db_ext import MibDbExternal, MibDbStatistic
from pyvoltha.adapters.common.kvstore.etcd_store import EtcdStore
from pyvoltha.common.utils.registry import registry


class AlarmDbExternal(MibDbExternal):
    """
    A persistent external OpenOMCI Alarm Database
    """
    CURRENT_VERSION = 1                       # VOLTHA v1.3.0 release
    ALARM_BITMAP_KEY = 'alarm_bit_map'

    _TIME_FORMAT = '%Y%m%d-%H%M%S.%f'

    ALARM_PATH = 'service/voltha/omci_alarms'
    DEVICE_PATH = '{}'  # .format(device_id)
    CLASS_PATH = DEVICE_PATH + '/classes/{}'  # .format(device_id, class_id)

    def __init__(self, omci_agent):
        """
        Class initializer
        :param omci_agent: (OpenOMCIAgent) OpenOMCI Agent
        """
        super(AlarmDbExternal, self).__init__(omci_agent)
        self._core = omci_agent.core_proxy
        # Some statistics to help with debug/tuning/...
        self._statistics = {
            'get': MibDbStatistic('get'),
            'set': MibDbStatistic('set'),
            'create': MibDbStatistic('create'),
            'delete': MibDbStatistic('delete')
        }
        self.args = registry('main').get_args()
        host, port = self.args.etcd.split(':', 1)
        self._kv_store = EtcdStore(host, port, AlarmDbExternal.ALARM_PATH)

    def _get_device_path(self, device_id):
        return AlarmDbExternal.DEVICE_PATH.format(device_id)

    def _get_class_path(self, device_id, class_id):
        if not self._started:
            raise DatabaseStateError('The Database is not currently active')

        if not 0 <= class_id <= 0xFFFF:
            raise ValueError('class-id is 0..0xFFFF')

        fmt = AlarmDbExternal.CLASS_PATH
        return fmt.format(device_id, class_id)

    def _time_to_string(self, time):
        return time.strftime(AlarmDbExternal._TIME_FORMAT) if time is not None else ''

    def _string_to_time(self, time):
        return datetime.strptime(time, AlarmDbExternal._TIME_FORMAT) if len(time) else None

    def _attribute_to_string(self, device_id, class_id, attr_name, value, old_value=None):
        """
        Convert an ME's attribute value to string representation

        :param value: (long) Alarm bitmaps are always a Long
        :return: (str) String representation of the value
        """
        self.log.debug("attribute-to-string", device_id=device_id, class_id=class_id,
                       attr_name=attr_name, value=value)
        return str(value)

    def _string_to_attribute(self, device_id, class_id, attr_name, str_value):
        """
        Convert an Alarm ME's attribute value-string to its Scapy decode equivalent

        :param device_id: (str) ONU Device ID
        :param class_id: (int) Class ID
        :param attr_name: (str) Attribute Name (at this point only alarm_bit_map)
        :param str_value: (str) Attribute Value in string form

        :return: (int)  Long integer representation of the value
        """
        # Alarms are always a bitmap which is a long
        if attr_name == AlarmDbExternal.ALARM_BITMAP_KEY:
            value = int(str_value) if len(str_value) else 0
        else:
            value = 0

        self.log.debug("string-to-attribute", device_id=device_id, class_id=class_id,
                       attr_name=attr_name, value=value)
        return value



#
# Copyright 2020 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import structlog
from pyvoltha.adapters.common.kvstore.twisted_etcd_store import TwistedEtcdStore
from pyvoltha.common.structlog_setup import setup_logging, update_logging, string_to_int
from twisted.internet.defer import inlineCallbacks, returnValue


COMPONENT_NAME = os.environ.get("COMPONENT_NAME")
GLOBAL_CONFIG_ROOT_NODE = "global"
DEFAULT_KV_STORE_CONFIG_PATH = "config"
KV_STORE_DATA_PATH_PREFIX = "service/voltha"
KV_STORE_PATH_SEPARATOR = "/"
CONFIG_TYPE = "loglevel"
DEFAULT_PACKAGE_NAME = "default"
GLOBAL_DEFAULT_LOGLEVEL = "WARN"

class LogController():
    instance_id = None
    active_log_level = None


    def __init__(self, etcd_host, etcd_port):
        self.log = structlog.get_logger()
        self.etcd_host = etcd_host
        self.etcd_port = etcd_port
        self.etcd_client = TwistedEtcdStore(self.etcd_host, self.etcd_port, KV_STORE_DATA_PATH_PREFIX)

    def make_config_path(self, key):
        return (DEFAULT_KV_STORE_CONFIG_PATH + KV_STORE_PATH_SEPARATOR + key + KV_STORE_PATH_SEPARATOR + CONFIG_TYPE + KV_STORE_PATH_SEPARATOR + DEFAULT_PACKAGE_NAME)


    @inlineCallbacks
    def get_global_loglevel(self):

        global_default_loglevel = ""

        try:
            level = yield self.etcd_client.get(self.global_config_path)
            if level is not None:
                level_int = string_to_int(str(level, 'utf-8'))

                if level_int == 0:
                    self.log.warn("Unsupported loglevel at global config path", level)
                else:
                    global_default_loglevel = level
                    self.log.debug("Retrieved global default loglevel", global_default_loglevel)

        except KeyError:
            self.log.warn("Failed to retrive default global loglevel")

        returnValue(global_default_loglevel)


    @inlineCallbacks
    def get_component_loglevel(self, global_default_loglevel):

        component_default_loglevel = global_default_loglevel

        try:
            level = yield self.etcd_client.get(self.component_config_path)
            if level is not None:
                level_int = string_to_int(str(level, 'utf-8'))

                if level_int == 0:
                    self.log.warn("Unsupported loglevel at component config path", level)

                else:
                    component_default_loglevel = level
                    self.log.debug("Retrieved component default loglevel", component_default_loglevel)

        except KeyError:
            self.log.warn("Failed to retrive default component loglevel")

        if component_default_loglevel == "":
            component_default_loglevel = GLOBAL_DEFAULT_LOGLEVEL.encode('utf-8')

        returnValue(component_default_loglevel)


    @inlineCallbacks
    def start_watch_log_config_change(self, instance_id, initial_default_loglevel):

        self.log.debug("Start watching for log config change")
        LogController.instance_id = instance_id

        if COMPONENT_NAME == None:
            raise Exception("Unable to retrive pod component name from runtime env")

        self.global_config_path = self.make_config_path(GLOBAL_CONFIG_ROOT_NODE)
        self.component_config_path = self.make_config_path(COMPONENT_NAME)

        self.set_default_loglevel(self.global_config_path, self.component_config_path, initial_default_loglevel.upper())
        self.process_log_config_change()

        yield self.etcd_client.watch(self.global_config_path, self.watch_callback)
        yield self.etcd_client.watch(self.component_config_path, self.watch_callback)


    def watch_callback(self, event):
        self.process_log_config_change()


    @inlineCallbacks
    def process_log_config_change(self):
        self.log.debug("Processing log config change")

        global_default_level = yield self.get_global_loglevel()
        level = yield self.get_component_loglevel(global_default_level)

        level_int = string_to_int(str(level, 'utf-8'))

        current_log_level = level_int
        if LogController.active_log_level != current_log_level:
            LogController.active_log_level = current_log_level
            self.log.debug("Applying updated loglevel")
            update_logging(LogController.instance_id, None, verbosity_adjust=level_int)

        else:
            self.log.debug("Loglevel not updated")


    @inlineCallbacks
    def set_default_loglevel(self, global_config_path, component_config_path, initial_default_loglevel):

        if (yield self.etcd_client.get(global_config_path)) == None:
            yield self.etcd_client.set(global_config_path, GLOBAL_DEFAULT_LOGLEVEL)

        if (yield self.etcd_client.get(component_config_path)) == None:
            yield self.etcd_client.set(component_config_path, initial_default_loglevel)
